import asyncio
from typing import List, Tuple, Coroutine

from .config import EventHandlerFunctor, EventFilter, E
from .filter import Filter, transaction_webhook_filter, bill_webhook_filter


class EventHandler:
    """
    Event handler, which executing and working with handlers

    """

    def __init__(self, functor: EventHandlerFunctor, *filter_: Filter) -> None:
        """

        :param functor:
        :param filter_:
        """
        self._fn = functor
        self._filters = filter_

    async def check_then_execute(self, event: E) -> Coroutine:
        """
        Check event, apply all filters and than pass on to handler

        :param event: handler
        """
        for filter_ in self._filters:
            if filter_.awaitable:
                if not await filter_.function(event):
                    break
            else:
                if not filter_.function(event):
                    break
        else:
            return await self._fn(event)


class HandlerManager:
    """
    Manager, which managing all handlers

    """

    def __init__(
            self,
            loop: asyncio.AbstractEventLoop
    ):
        if not isinstance(loop, asyncio.AbstractEventLoop):
            raise ValueError(
                f"Listener must have its event loop implemented with"
                f" {asyncio.AbstractEventLoop!r}"
            )

        self.loop = loop
        self.transaction_handlers: List[EventHandler] = []
        self.bill_handlers: List[EventHandler] = []

    @property
    def handlers(self) -> List[EventHandler]:
        """ Return all registered handlers """
        return [*self.bill_handlers, *self.transaction_handlers]

    @staticmethod
    def wrap_handler(
            event_handler: EventHandlerFunctor,
            filters: Tuple[EventFilter, ...],
            default_filter: Filter
    ) -> EventHandler:
        """
        Add new event handler.
        (!) Allows chain addition.
        :param event_handler: event handler, low order function
         which works with events
        :param filters: filter for low order function execution
        :param default_filter: default filter for handler
        :return: this handler manager
        """
        if filters:  # Initially filters are in tuple
            filters_list = list(filters)

            for index, filter_ in enumerate(filters):
                if not isinstance(filter_, Filter):
                    filters_list[index] = default_filter & Filter(filter_)

            filters = tuple(filters_list)
        else:
            filters = (default_filter,)

        return EventHandler(event_handler, *filters)

    def add_transaction_handler(self, *filters: EventFilter) -> E:

        def decorator(event_handler: EventHandlerFunctor) -> None:
            self.transaction_handlers.append(
                self.wrap_handler(
                    event_handler, filters,
                    default_filter=transaction_webhook_filter
                )
            )

        return decorator

    def add_bill_handler(self, *filters: EventFilter) -> E:

        def decorator(event_handler: EventHandlerFunctor) -> None:
            self.bill_handlers.append(
                self.wrap_handler(
                    event_handler, filters,
                    default_filter=bill_webhook_filter
                )
            )

        return decorator

    async def process_event(self, event: E) -> None:
        """
        Feed handlers with event.
        :param event: any object that will be translated to handlers
        """

        for handler in self.handlers:
            await handler.check_then_execute(event)



from .account import Account
from .account_info import QiwiAccountInfo
from .balance import Balance
from .bill import Bill, BillError, RefundBill, Notification
from .identification import Identification
from .limit import Limit
from .partner import Partner
from .payment_info import PaymentInfo
from .polygon import Polygon
from .qiwi_master import OrderDetails
from .stats import Statistic
from .terminal import Terminal
from .transaction import Transaction
from .webhooks import WebHookConfig, WebHook

__all__ = (
    'QiwiAccountInfo',
    'Transaction',
    'Bill',
    'BillError',
    'Statistic',
    'Limit',
    'Account',
    'Balance',
    'Identification',
    'PaymentInfo',
    'OrderDetails',
    'RefundBill',
    'Polygon',
    'Terminal',
    'Partner',
    'WebHook',
    'WebHookConfig',
    'Notification'
)



"""Main model: Partner"""
from typing import List, Optional

from pydantic import BaseModel

from glQiwiApi.utils.basics import custom_load


class Partner(BaseModel):
    """ Base partner class """
    title: str
    id: int

    maps: Optional[List[str]] = None

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = ("Partner",)



from typing import Optional

from pydantic import BaseModel, Field

from glQiwiApi.types import Sum


class OrderDetails(BaseModel):
    """ object: OrderDetails """
    order_id: str = Field(..., alias="id")
    card_alias: str = Field(..., alias="cardAlias")
    status: str
    price: Optional[Sum] = None
    card_id: Optional[str] = Field(alias="cardId", default=None)


__all__ = (
    'OrderDetails'
)



from typing import Optional

from pydantic import BaseModel, Field

from glQiwiApi.types.basics import Sum, Type
from glQiwiApi.utils.basics import custom_load


class Account(BaseModel):
    """ object: Account """
    alias: str
    title: str
    fs_alias: str = Field(alias="fsAlias")
    bank_alias: str = Field(alias="bankAlias")
    has_balance: bool = Field(alias="hasBalance")
    balance: Optional[Sum] = Field(const=None)
    currency: int
    account_type: Optional[Type] = Field(None, alias="type")
    is_default_account: bool = Field(alias="defaultAccount")

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = [
    'Account'
]



from aiohttp.abc import AbstractAccessLogger


class AccessLogger(AbstractAccessLogger):
    """ Custom logger for aiohttp.web.Application """

    def log(self, request, response, time) -> None:
        self.logger.info(f'{request.remote} '
                         f'"{request.method} {request.path} '
                         f'done in {time}s: {response.status}')



from typing import Optional

from pydantic import BaseModel, Field

from glQiwiApi.types import Sum
from glQiwiApi.utils.basics import custom_load


class Fields(BaseModel):
    """ Специальные поля """
    account: str


class State(BaseModel):
    """ State """
    code: str


class TransactionInfo(BaseModel):
    """ Информация о транзакции """
    txn_id: int = Field(..., alias="id")
    state: State


class PaymentInfo(BaseModel):
    """Информация о платеже"""
    payment_id: int = Field(..., alias="id")
    terms: str
    fields: Fields
    payment_sum: Sum = Field(..., alias="sum")
    source: str
    transaction: Optional[TransactionInfo] = None
    comment: Optional[str] = None

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = [
    'PaymentInfo'
]



from datetime import datetime
from enum import Enum
from typing import Optional, Union, Dict, List, Any

from pydantic import BaseModel, Field, Extra

from glQiwiApi.utils.basics import custom_load


class BalanceDetails(BaseModel):
    """ object: BalanceDetails """
    total: float
    available: float
    deposition_pending: Optional[float] = None
    blocked: Optional[float] = None
    debt: Optional[float] = None
    hold: Optional[float] = None


class CardsLinked(BaseModel):
    """ object: CardsLinked """
    pan_fragment: str
    type: str


class AccountInfo(BaseModel):
    """ object: AccountInfo"""

    account: str
    """Номер счета"""

    balance: float
    """Баланс счета"""

    currency: str
    """
    Код валюты счета пользователя.
    Всегда 643 (рубль РФ по стандарту ISO 4217).
    """

    identified: bool
    """Верифицирован ли кошелек"""

    account_type: str
    """
    Тип счета пользователя. Возможные значения:
    personal — счет пользователя в ЮMoney;
    professional — профессиональный счет в ЮMoney
    """

    account_status: str
    """
    Статус пользователя. Возможные значения:
    anonymous — анонимный счет;
    named — именной счет;
    identified — идентифицированный счет.
    """

    balance_details: BalanceDetails
    """
    Расширенная информация о балансе.
    По умолчанию этот блок отсутствует.
    Блок появляется, если сейчас или когда-либо ранее были зачисления в очереди
    задолженности, блокировки средств.
    """

    cards_linked: Optional[List[CardsLinked]] = None
    """
    Информация о привязанных банковских картах.
    Если к счету не привязано ни одной карты, параметр отсутствует.
    Если к счету привязана хотя бы одна карта,
    параметр содержит список данных о привязанных картах.
    pan_fragment - string Маскированный номер карты.
    type: string
    Тип карты. Может отсутствовать, если неизвестен. Возможные значения:
    - VISA;
    - MasterCard;
    - AmericanExpress;
    - JCB.
    """

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


class OperationType(Enum):
    """
    Типы операций YooMoney

    deposition — пополнение счета (приход);
    payment — платежи со счета (расход);
    incoming_transfers_unaccepted —
    непринятые входящие P2P-переводы любого типа.

    """
    DEPOSITION = 'deposition'
    """Пополнение счета (приход);"""

    PAYMENT = 'payment'
    """Платежи со счета (расход)"""

    TRANSFERS = 'incoming-transfers-unaccepted'
    """непринятые входящие P2P-переводы любого типа."""


class DigitalGoods(BaseModel):
    """
    Данные о цифровом товаре (пин-коды и бонусы игр, iTunes, Xbox, etc.)
    Поле присутствует при успешном платеже в магазины цифровых товаров.
    Описание формата:
    https://yoomoney.ru/docs/wallet/process-payments/process-payment#digital-goods
    """
    article_id: str = Field(..., alias="merchantArticleId")
    serial: str
    secret: str


class Operation(BaseModel):
    """ object: Operation """

    operation_id: str
    """Идентификатор операции."""

    status: str
    """
    Статус платежа (перевода). Может принимать следующие значения:
    - success — платеж завершен успешно;
    - refused — платеж отвергнут получателем или отменен отправителем;
    - in_progress — платеж не завершен,
     перевод не принят получателем или ожидает ввода кода протекции.
    """

    operation_date: datetime = Field(..., alias="datetime")
    """
    Дата и время совершения операции в формате строки
    в ISO формате с часовым поясом UTC.
    """

    title: str
    """
    Краткое описание операции (название магазина или источник пополнения).
    """

    direction: str
    """
    Направление движения средств. Может принимать значения:
    - in (приход);
    - out (расход).
    """

    amount: Union[int, float]
    """Сумма операции."""

    operation_type: str = Field(..., alias="type")
    """Тип операции. Возможные значения:
    payment-shop — исходящий платеж в магазин;
    outgoing-transfer — исходящий P2P-перевод любого типа;
    deposition — зачисление;
    incoming-transfer — входящий перевод или перевод до востребования;
    incoming-transfer-protected — входящий перевод с кодом протекции.
    """

    label: Optional[str] = None
    """Метка платежа.
     Присутствует для входящих и исходящих переводов другим пользователям
     у которых был указан параметр label вызова send()
     """

    pattern_id: Optional[str] = None
    """
    Идентификатор шаблона, по которому совершен платеж.
    Присутствует только для платежей.
    """

    details: Optional[Any] = None

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


class OperationDetails(BaseModel):
    """ object: OperationDetails """
    operation_id: Optional[str] = None
    """Идентификатор операции. Можно получить при вызове метода history()"""

    status: Optional[str] = None
    """Статус платежа (перевода). Можно получить при вызове метода history()"""

    amount: Optional[float] = None
    """Сумма операции (сумма списания со счета)."""

    operation_date: Optional[str] = Field(..., alias="datetime")
    """
    Дата и время совершения операции в формате строки
    в ISO формате с часовым поясом UTC.
    """

    operation_type: Optional[str] = Field(..., alias="type")
    """Тип операции. Возможные значения:
    payment-shop — исходящий платеж в магазин;
    outgoing-transfer — исходящий P2P-перевод любого типа;
    deposition — зачисление;
    incoming-transfer — входящий перевод или перевод до востребования;
    incoming-transfer-protected — входящий перевод с кодом протекции.
    """

    direction: Optional[str] = None
    """
    направление движения средств. может принимать значения:
    - in (приход);
    - out (расход).
    """

    comment: Optional[str] = None
    """
    Комментарий к переводу или пополнению.
    Присутствует в истории отправителя перевода или получателя пополнения.
    """

    digital_goods: Optional[Dict[str, DigitalGoods]] = None
    """
    Данные о цифровом товаре (пин-коды и бонусы игр, iTunes, Xbox, etc.)
    Поле присутствует при успешном платеже в магазины цифровых товаров.
    Описание формата:
    https://yoomoney.ru/docs/wallet/process-payments/process-payment#digital-goods
    """

    details: Optional[str] = None
    """
    Детальное описание платежа. Строка произвольного формата,
    может содержать любые символы и переводы строк
    Необязательный параметр.
    """

    label: Optional[str] = None
    """Метка платежа."""

    answer_datetime: Optional[str] = None
    """
    Дата и время приема или отмены перевода, защищенного кодом протекции.
    Присутствует для входящих и исходящих переводов, защищенных кодом протекции
    если при вызове send вы указали protect=True при передаче аргументов.
    Если перевод еще не принят или не отвергнут получателем, поле отсутствует.
    """

    expires: Optional[str] = None
    """
    Дата и время истечения срока действия кода протекции.
    Присутствует для входящих и исходящих переводов (от/другим) пользователям,
    защищенных кодом протекции,
    если при вызове send вы указали protect=True при передаче аргументов.
    """

    protection_code: Optional[str] = None
    """
    Код протекции.
    Присутствует для исходящих переводов, защищенных кодом протекции.
    """

    codepro: Optional[bool] = None
    """
    Перевод защищен кодом протекции.
    Присутствует для переводов другим пользователям.
    """

    message: Optional[str] = None
    """
    Сообщение получателю перевода.
    Присутствует для переводов другим пользователям.
    """

    recipient_type: Optional[str] = None
    """
    Тип идентификатора получателя перевода. Возможные значения:
    account — номер счета получателя в сервисе ЮMoney;
    phone — номер привязанного мобильного телефона получателя;
    email — электронная почта получателя перевода.
    Присутствует для исходящих переводов другим пользователям.
    """

    recipient: Optional[str] = None
    """
    Идентификатор получателя перевода.
    Присутствует для исходящих переводов другим пользователям.
    """

    sender: Optional[str] = None
    """
    Номер счета отправителя перевода.
    Присутствует для входящих переводов от других пользователей.
    """

    title: Optional[str] = None
    """
    Краткое описание операции (название магазина или источник пополнения).
    """

    fee: Optional[float] = None
    """
    Сумма комиссии.
    Присутствует для исходящих переводов другим пользователям.
    """

    amount_due: Optional[float] = None
    """
    Сумма к получению.
    Присутствует для исходящих переводов другим пользователям.
    """

    pattern_id: Optional[str] = None
    """
    Идентификатор шаблона платежа, по которому совершен платеж.
    Присутствует только для платежей.
    """

    error: Optional[str] = None
    """
    Код ошибки, присутствует при ошибке выполнения запрос
    Возможные ошибки:
    illegal_param_operation_id - неверное значение параметра operation_id
    Все прочие значения - техническая ошибка, повторите вызов метода позднее.
    """

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


class Wallet(BaseModel):
    """ object: Wallet """
    allowed: bool


class Item(BaseModel):
    """ object: Item """
    item_id: str = Field(..., alias="id")
    """
    Идентификатор привязанной к счету банковской карты.
    Его необходимо указать в методе process-payment для
    совершения платежа выбранной картой.
    """
    pan_fragment: str
    """
    Фрагмент номера банковской карты.
    Поле присутствует только для привязанной банковской карты.
    Может отсутствовать, если неизвестен.
    """
    item_type: str = Field(..., alias="type")
    """
    Тип карты. Может отсутствовать, если неизвестен. Возможные значения:
    Visa;
    MasterCard;
    American Express;
    JCB.
    """


class Card(BaseModel):
    """ object: Card"""
    allowed: bool
    csc_required: bool
    items: List[Item]


class MoneySource(BaseModel):
    """
    Список доступных методов для проведения данного платежа.
    Каждый метод содержит набор атрибутов.
    """
    wallet: Wallet
    """платеж со счета пользователя"""
    cards: Optional[Card] = None
    """платеж с банковских карт, привязанных к счету"""


class PreProcessPaymentResponse(BaseModel):
    """
    Объект, который вы получаете при вызове _pre_process_payment.
    При вызове данного метода вы не списываете деньги со своего счёта,
    а условно подготавливаете его к отправке.
    Для отправки денег на счёт используйте метод send()
    """
    status: str
    request_id: str
    recipient_account_status: str
    fees: Dict[str, float]
    balance: Optional[float] = None
    recipient_account_type: Optional[str] = None
    recipient_identified: bool = False
    recipient_masked_account: Optional[str] = None
    multiple_recipients_found: Optional[str] = None
    contract_amount: Optional[float] = None
    error: Optional[str] = None
    money_source: Optional[MoneySource] = None
    protection_code: Optional[str] = None
    account_unblock_uri: Optional[str] = None
    ext_action_uri: Optional[str] = None

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


class Payment(BaseModel):  # lgtm [py/missing-equals #
    """ object: Payment """
    status: str
    """
    Код результата выполнения операции. Возможные значения:
    success — успешное выполнение (платеж проведен).
    Это конечное состояние платежа.
    refused — отказ в проведении платежа.
    Причина отказа возвращается в поле error. Это конечное состояние платежа.
    in_progress — авторизация платежа не завершена.
    Приложению следует повторить запрос
    с теми же параметрами спустя некоторое время.
    ext_auth_required — для завершения авторизации платежа
    с использованием банковской карты
    требуется аутентификация по технологии 3‑D Secure.
    все прочие значения — состояние платежа неизвестно.
    Приложению следует повторить запрос с теми же параметрами
    спустя некоторое время.
    """

    payment_id: str
    """
    Идентификатор проведенного платежа.
    Присутствует только при успешном выполнении метода send().
    """

    credit_amount: Optional[float] = None
    """
    Сумма, поступившая на счет получателя.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payer: Optional[str] = None
    """
    Номер счета плательщика.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payee: Optional[str] = None
    """
    Номер счета получателя.
    Присутствует при успешном переводе средств
    на счет другого пользователя ЮMoney.
    """

    payee_uid: Union[str, int, None] = None
    """Служебное значение, не фигурирует в документации"""

    invoice_id: Optional[str] = None
    """
    Номер транзакции магазина в ЮMoney.
    Присутствует при успешном выполнении платежа в магазин.
    """

    balance: Optional[float] = None
    """
    Баланс счета пользователя после проведения платежа.
    Присутствует только при выполнении следующих условий:
    - метод выполнен успешно;
    - токен авторизации обладает правом account-info.
    """

    error: Optional[str] = None
    """
    Код ошибки при проведении платежа (пояснение к полю status).
    Присутствует только при ошибках.
    """
    account_unblock_uri: Optional[str] = None
    """
    Адрес, на который необходимо отправить пользователя для разблокировки счета
    Поле присутствует в случае ошибки account_blocked.
    """

    acs_uri: Optional[str] = None

    acs_params: Optional[str] = None

    next_retry: Optional[int] = None
    """
    Рекомендуемое время,
    спустя которое следует повторить запрос, в миллисекундах.
    Поле присутствует при status=in_progress.
    """

    digital_goods: Optional[Dict[str, Dict[str, List[DigitalGoods]]]] = None

    protection_code: Optional[str] = None
    """
    Код протекции, который был сгенерирован,
    если при вызове метода апи send вы указали protect=True
    при передаче аргументов
    """

    def initialize(self, protection_code: str) -> 'Payment':
        self.protection_code = protection_code
        return self

    class Config:
        """ Pydantic config """
        json_loads = custom_load
        extra = Extra.allow

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


class IncomingTransaction(BaseModel):
    """ object: IncomingTransaction """
    status: str
    protection_code_attempts_available: int
    ext_action_uri: Optional[str] = None
    error: Optional[str] = None

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = (
    'AccountInfo',
    'OperationType',
    'Operation',
    'OperationDetails',
    'PreProcessPaymentResponse',
    'Payment',
    'IncomingTransaction'
)



from typing import List

from pydantic import BaseModel, Field

from glQiwiApi.types.basics import Sum
from glQiwiApi.utils.basics import custom_load


class Statistic(BaseModel):
    """ object: Statistic """
    incoming: List[Sum] = Field(alias="incomingTotal")
    out: List[Sum] = Field(alias="outgoingTotal")

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = [
    'Statistic'
]



class Polygon:
    """ Polygon class for QiwiMaps class """

    def __init__(self, lat_lon_pair_nw: tuple, lat_lon_pair_se: tuple):
        self.lat_nw, self.lon_nw = lat_lon_pair_nw
        self.lat_se, self.lon_se = lat_lon_pair_se

        self._dict = {
            "latNW": self.lat_nw,
            "lngNW": self.lon_nw,
            "latSE": self.lat_se,
            "lngSE": self.lon_se,
        }

    @property
    def dict(self):
        return {k: str(double) for k, double in self._get_items()}

    def _get_items(self):
        return self._dict.items()


__all__ = ("Polygon",)



from datetime import datetime
from typing import Union

from pydantic import BaseModel, Field

from glQiwiApi.utils.basics import custom_load


class Interval(BaseModel):
    """ object: Interval """
    date_from: datetime = Field(alias="dateFrom")
    date_till: datetime = Field(alias="dateTill")


class Limit(BaseModel):
    """ object: Limit """
    currency: str
    rest: Union[float, int]
    max_limit: Union[float, int] = Field(alias="max")
    spent: Union[float, int]
    interval: Interval
    limit_type: str = Field(alias="type")

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()


__all__ = [
    'Limit'
]



from pydantic import BaseModel

from glQiwiApi.utils.basics import custom_load


class Balance(BaseModel):
    """ object: Balance """
    alias: str
    currency: int

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return f'Config class with loads={self.json_loads}'

        def __repr__(self) -> str:
            return self.__str__()



import copy
from typing import Optional, Any


class BillMixin(object):
    """
    Примесь, позволяющая проверять счет, не используя метод QiwiWrapper,
    добавляя метод check() объекту Bill

    """
    _w = None
    bill_id: Optional[str] = None

    def initialize(self, wallet: Any):
        self._w = copy.copy(wallet)
        return self

    async def check(self) -> bool:
        async with self._w:
            return (await self._w.check_p2p_bill_status(
                bill_id=self.bill_id
            )) == 'PAID'


class ToolsMixin(object):
    """ Object: ToolsMixin """
    _requests = None

    async def __aenter__(self):
        """Создаем сессию, чтобы не пересоздавать её много раз"""
        self._requests.create_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Закрываем сессию и очищаем кэш при выходе"""
        if self._requests.session:
            await self._requests.session.close()
            self._requests.clear_cache()

    def _get(self, item: Any) -> Any:
        try:
            return super().__getattribute__(item)
        except AttributeError:
            return None

    def __deepcopy__(self, memo) -> 'ToolsMixin':
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        dct = {slot: self._get(slot) for slot in self.__slots__ if
               self._get(slot) is not None}
        for k, value in dct.items():
            if k == '_parser':
                value.session = None
            setattr(result, k, copy.deepcopy(value, memo))

        return result



from typing import Dict, Optional, Any, Union

import aiohttp

import glQiwiApi
from glQiwiApi.core.basic_requests_api import HttpXParser
from glQiwiApi.core.storage import Storage
from glQiwiApi.types import Response
from glQiwiApi.types.basics import Cached
from glQiwiApi.utils.exceptions import RequestError


class RequestManager(HttpXParser):
    """
    Немного переделанный наследник HttpXParser
    под платежные системы и кэширование запросов

    """
    __slots__ = ('without_context', 'messages', '_cache', '_cached_key')

    def __init__(
            self,
            without_context: bool = False,
            messages: Optional[Dict[str, str]] = None,
            cache_time: Union[float, int] = 0
    ) -> None:
        super(RequestManager, self).__init__()
        self.without_context = without_context
        self.messages = messages
        self._cache = Storage(cache_time)
        self._cached_key = "session"

    def clear_cache(self) -> None:
        """ Clear all cache in storage """
        self._cache.clear(force=True)

    def get_cached_session(self) -> Optional[aiohttp.ClientSession]:
        """ Get cached session from storage """
        cached = self._cache[self._cached_key]
        if self.check_session(cached):
            return cached
        return None

    def set_cached_session(self):
        cached_session = self.get_cached_session()
        if cached_session:
            self.session = cached_session

    async def _request(self, *args, **kwargs) -> Response:
        # Получаем текущий кэш используя ссылку как ключ
        response = self._cache.get_current(kwargs.get('url'))
        if not self._cache.validate(kwargs):
            self.set_cached_session()
            response = await super()._request(*args, **kwargs)
        # Проверяем, не был ли запрос в кэше, если нет,
        # то проверяем статус код и если он не 200 - выбрасываем ошибку
        if not isinstance(response, Cached):
            if response.status_code != 200:
                await self._close_session()
                self.raise_exception(
                    response.status_code,
                    json_info=response.response_data
                )
            await self._close_session()

        self._cache_all(response, kwargs)

        return response

    def raise_exception(
            self,
            status_code: Union[str, int],
            json_info: Optional[Dict[str, Any]] = None,
            message: Optional[str] = None
    ) -> None:
        """ Raise RequestError exception with pretty explanation """
        if not isinstance(message, str):
            message = self.messages.get(str(status_code), "Unknown")
        raise RequestError(
            message,
            status_code,
            additional_info=f"{glQiwiApi.__version__} version api",
            json_info=json_info
        )

    def cache_session(self, session: aiohttp.ClientSession) -> None:
        if self.check_session(session):
            self._cache.update_data(
                obj_to_cache=session,
                key='session'
            )

    async def _close_session(self) -> None:
        if self.without_context:
            await self.session.close()

    def _cache_all(self, response: Response, kwargs: Dict[Any, Any]):
        resolved = self._cache.initialize_response_to_resolve(
            result=response.response_data,
            kwargs=kwargs,
            status_code=response.status_code
        )
        self._cache.update_data(
            key=None,
            obj_to_cache=resolved
        )
        self.cache_session(self.session)

    @staticmethod
    def check_session(session: Any) -> bool:
        if isinstance(session, aiohttp.ClientSession):
            if not session.closed:
                return True
        return False

    def create_session(self, **kwargs) -> None:
        """ Create new session or get it from cache """
        self.set_cached_session()
        super().create_session(**kwargs)

    @classmethod
    def filter_dict(cls, dictionary: dict) -> dict:
        """
        Pop NoneType values and convert everything to str, designed?for=params
        :param dictionary: source dict
        :return: filtered dict
        """
        return {k: str(v) for k, v in dictionary.items() if v is not None}



import time

from glQiwiApi.types import WrapperData

BASE_QIWI_URL = 'https://edge.qiwi.com'
BASE_P2P_URL = 'https://api.qiwi.com/partner/bill/v1/bills/'

DEFAULT_QIWI_HEADERS = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
    'Authorization': 'Bearer {token}',
    'Host': 'edge.qiwi.com',
}

ERROR_CODE_NUMBERS = {
    "400": "Недостаточно средств для проведения операции",
    "401": "Неверный токен или истек срок действия токена API",
    "403": "Нет прав на данный запрос"
           "(недостаточно разрешений у токена API)",
    "404": "Не найдена транзакция или "
           "отсутствуют платежи с указанными признаками",
    "423": "Слишком много запросов, сервис временно недоступен",
    "422": "Неправильно указаны домен/подсеть/хост"
           " веб-хука(в параметре new_url для URL веб-хука), "
           "неправильно указаны тип хука или тип транзакции, "
           "попытка создать хук при наличии уже созданного",
    "405": "Ошибка, связанная с типом запроса к апи,"
           "обратитесь к разработчику или откройте issue",
    "500": "Внутренняя ошибка сервиса",
    "400_special_bad_proxy": "Ошибка, связанная с использованием прокси"
}

P2P_QIWI_HEADERS = {
    'Authorization': 'Bearer {token}',
    'Accept': 'application/json',
    'Content-Type': 'application/json',
}

QIWI_TO_CARD = WrapperData(
    json={
        "id": str(int(time.time() * 1000)),
        "sum": {
            "amount": "", "currency": "643"
        },
        "paymentMethod": {
            "type": "Account", "accountId": "643"
        },
        "fields": {
            "account": ""
        }
    },
    headers=DEFAULT_QIWI_HEADERS
)

P2P_DATA = WrapperData(
    json={
        "amount": {
            "currency": "RUB",
            "value": "{amount}"
        },
        "expirationDateTime": "",
        "comment": "{comment}",
        "customFields": {
            "paySourcesFilter": "qw",
            "themeCode": "Yvan-YKaSh",
        }
    },

    headers=P2P_QIWI_HEADERS
)

QIWI_TO_WALLET = WrapperData(
    json={
        "id": str(int(time.time() * 1000)),
        "sum": {
            "amount": "",
            "currency": ""
        },
        "paymentMethod": {
            "type": "Account",
            "accountId": "643"
        },
        "comment": "",
        "fields": {
            "account": ""
        }
    },
    headers=DEFAULT_QIWI_HEADERS
)

TRANSACTION_TRANSFER = {
    'txnId': 'transaction_id',
    'personId': 'person_id',
    'account': 'to_account'
}

IDENTIFICATION_TRANSFER = {
    'id': 'identification_id',
    'firstName': 'first_name',
    'middleName': 'middle_name',
    'lastName': 'last_name',
    'birthDate': 'birth_date'
}

P2P_BILL_TRANSFER = {
    'siteId': 'site_id',
    'billId': 'bill_id',
    'creationDateTime': 'creation_date_time',
    'expirationDateTime': 'expiration_date_time',
    'payUrl': 'pay_url',
    'customFields': 'custom_fields',
    'value': 'amount'
}

LIMIT_TYPES = [
    'TURNOVER',
    'REFILL',
    'PAYMENTS_P2P',
    'PAYMENTS_PROVIDER_INTERNATIONALS',
    'PAYMENTS_PROVIDER_PAYOUT',
    'WITHDRAW_CASH'
]
LIMIT_TYPES_TRANSFER = {
    'max': 'max_limit',
    'type': 'limit_type'
}
COMMISSION_DATA = WrapperData(
    json={
        "account": "",
        "paymentMethod": {
            "type": "Account",
            "accountId": "643"
        },
        "purchaseTotals":
            {
                "total":
                    {"amount": "",
                     "currency": "643"
                     }
            }},
    headers=DEFAULT_QIWI_HEADERS
)

COMMISSION_TRANSFER = {
    'providerId': 'provider_id',
    'withdrawSum': 'withdraw_sum',
    'qwCommission': 'qw_commission',
    'withdrawToEnrollmentRate': 'withdraw_to_enrollment_rate'
}

__all__ = (
    'LIMIT_TYPES',
    'TRANSACTION_TRANSFER',
    'IDENTIFICATION_TRANSFER',
    'QIWI_TO_WALLET',
    'QIWI_TO_CARD',
    'DEFAULT_QIWI_HEADERS',
    'LIMIT_TYPES_TRANSFER',
    'P2P_QIWI_HEADERS',
    'P2P_DATA',
    'P2P_BILL_TRANSFER',
    'COMMISSION_DATA',
    'COMMISSION_TRANSFER',
    'ERROR_CODE_NUMBERS',
    'BASE_QIWI_URL',
    'BASE_P2P_URL'
)



import ipaddress
import logging
import typing

from aiohttp import web
from aiohttp.web_response import Response

from glQiwiApi import types
from glQiwiApi.core.abstracts import BaseWebHookView
from glQiwiApi.core.web_hooks.handler import HandlerManager
from glQiwiApi.utils.basics import hmac_for_transaction, hmac_key
from .config import (
    DEFAULT_QIWI_WEBHOOK_PATH,
    allowed_ips,
    DEFAULT_QIWI_ROUTER_NAME,
    DEFAULT_QIWI_BILLS_ROUTER_NAME,
    DEFAULT_QIWI_BILLS_WEBHOOK_PATH,
    Path
)


def _check_ip(ip_address: str) -> bool:
    """
    Check if ip is allowed to request us
    :param ip_address: IP-address
    :return: address is allowed
    """
    address = ipaddress.IPv4Address(ip_address)
    unpacked = [ip_addr for pool in allowed_ips for ip_addr in pool]
    return address in unpacked


class QiwiWalletWebView(BaseWebHookView):
    """
    View, which processes transactions

    """

    def _check_ip(self, ip_address: str) -> bool:
        return _check_ip(ip_address)

    async def parse_update(self) -> types.WebHook:
        """
        Deserialize update and create new update class
        :return: :class:`updated.QiwiUpdate`
        """
        data = await self.request.json()
        return types.WebHook.parse_raw(data)

    async def post(self) -> Response:
        await super().post()
        return web.Response(text="ok", status=200)

    def _hash_validator(self, update: types.WebHook) -> None:
        base64_key = self.request.app.get('_base64_key')

        if update.test:
            return

        validated = hmac_for_transaction(
            webhook_key_base64=base64_key,
            amount=update.payment.sum,
            txn_hash=update.hash,
            txn_type=update.payment.type,
            txn_id=update.payment.txn_id,
            account=update.payment.account
        )

        if not validated:
            raise web.HTTPBadRequest()

    app_key_check_ip = "_qiwi_wallet_check_ip"
    app_key_handler_manager = "_qiwi_wallet_handler_manager"


class QiwiBillWebView(BaseWebHookView):
    """
    View, which processes p2p notifications


    """

    def _check_ip(self, ip_address: str) -> bool:
        return _check_ip(ip_address)

    def _hash_validator(self, update: types.Notification) -> None:
        sha256_signature = self.request.headers.get("X-Api-Signature-SHA256")
        logging.info(sha256_signature)
        _secret = self.request.app.get("_secret_key")
        answer = hmac_key(_secret, update.bill.amount,
                          update.bill.status, update.bill.bill_id,
                          update.bill.site_id)

        if answer != sha256_signature:
            raise web.HTTPBadRequest()

    async def parse_update(self) -> types.Notification:
        payload = await self.request.json()
        return types.Notification.parse_raw(payload)

    async def post(self) -> Response:
        self.validate_ip()

        update = await self.parse_update()

        # self._hash_validator(update)

        await self.handler_manager.process_event(update)
        return web.json_response(data={"error": "0"}, status=200)

    app_key_check_ip = "_qiwi_bill_check_ip"
    app_key_handler_manager = "_qiwi_bill_handler_manager"


def setup_transaction_data(
        app: web.Application,
        base64_key: str,
        handler_manager: HandlerManager,
        path: typing.Optional[str] = None
):
    app["_base64_key"] = base64_key
    app[QiwiWalletWebView.app_key_check_ip] = _check_ip
    app[QiwiWalletWebView.app_key_handler_manager] = handler_manager
    path = path or DEFAULT_QIWI_WEBHOOK_PATH
    app.router.add_view(path, QiwiWalletWebView, name=DEFAULT_QIWI_ROUTER_NAME)


def setup_bill_data(
        app: web.Application,
        secret_key: str,
        handler_manager: HandlerManager,
        path: typing.Optional[str] = None
) -> None:
    app["_secret_key"] = secret_key
    app[QiwiBillWebView.app_key_check_ip] = _check_ip
    app[QiwiBillWebView.app_key_handler_manager] = handler_manager
    app.router.add_view(
        handler=QiwiBillWebView,
        name=DEFAULT_QIWI_BILLS_ROUTER_NAME,
        path=path or DEFAULT_QIWI_BILLS_WEBHOOK_PATH
    )


def setup(
        handler_manager: HandlerManager,
        app: web.Application,
        path: typing.Optional[Path] = None,
        secret_key: typing.Optional[str] = None,
        base64_key: typing.Optional[str] = None
) -> None:
    """
    Setup web application for webhooks

    :param handler_manager: handler_manager, which processing events
    :param app: aiohttp.web.Application
    :param path: Path obj, contains two paths
    :param secret_key: secret p2p key
    :param base64_key: Base64-encoded webhook key
    """
    setup_bill_data(app, secret_key, handler_manager,
                    path.bill_path)
    setup_transaction_data(app, base64_key, handler_manager,
                           path.transaction_path)



from datetime import datetime
from typing import List, Dict, Any, Union, Optional, Tuple

from pydantic import ValidationError

import glQiwiApi.utils.basics as api_helper
from glQiwiApi.core import (
    AbstractPaymentWrapper,
    RequestManager,
    ToolsMixin
)
from glQiwiApi.types import (
    AccountInfo,
    OperationType,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    IncomingTransaction
)
from glQiwiApi.types.basics import DEFAULT_CACHE_TIME
from glQiwiApi.utils.exceptions import NoUrlFound, InvalidData
from glQiwiApi.yoo_money.basic_yoomoney_config import (
    BASE_YOOMONEY_URL,
    ERROR_CODE_NUMBERS,
    content_and_auth
)


class YooMoneyAPI(AbstractPaymentWrapper, ToolsMixin):
    """
    Класс, реализующий обработку запросов к YooMoney
    Удобен он тем, что не просто отдает json подобные объекты,
    а всё это конвертирует в python dataclasses.
    Для работы с данным классом, необходимо зарегистрировать токен,
    используя гайд на официальном github проекта

    """

    __slots__ = ("api_access_token", "_requests",)

    def __init__(
            self,
            api_access_token: str,
            without_context: bool = False,
            cache_time: Union[float, int] = DEFAULT_CACHE_TIME
    ) -> None:
        """
        Конструктор принимает токен, полученный из класс метода
         get_access_token
        и специальный аттрибут without_context,

        :param api_access_token: апи токен для запросов
        :param without_context: bool, указывающая будет
         ли объект класса "глобальной" переменной
         или будет использована в async with контексте
        :param cache_time: Время кэширование запросов в секундах,
         по умолчанию 0, соответственно,
         запрос не будет использовать кэш по дефолту, максимальное время
         кэширование 60 секунд
        """
        self.api_access_token = api_access_token
        self._requests = RequestManager(
            without_context=without_context,
            messages=ERROR_CODE_NUMBERS,
            cache_time=cache_time
        )

    def _auth_token(self, headers: dict) -> Dict[Any, Any]:
        headers['Authorization'] = headers['Authorization'].format(
            token=self.api_access_token
        )
        return headers

    @classmethod
    async def build_url_for_auth(
            cls,
            scope: List[str],
            client_id: str,
            redirect_uri: str = 'https://example.com'
    ) -> str:
        """
        Метод для получения ссылки для
        дальнейшей авторизации и получения токена

        :param scope: OAuth2-авторизации приложения пользователем,
         права передаются списком.
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда попадет временный код, который нужен
         для получения основного токена
        :return: ссылку, по которой нужно перейти
         и сделать авторизацию через логин/пароль
        """
        headers = api_helper.parse_headers()
        params = {
            'client_id': client_id,
            'response_type': 'code',
            'redirect_uri': redirect_uri,
            'scope': " ".join(scope)
        }
        async for response in RequestManager(
                without_context=True,
                messages=ERROR_CODE_NUMBERS,
                cache_time=DEFAULT_CACHE_TIME
        ).fast().fetch(
            url=BASE_YOOMONEY_URL + '/oauth/authorize',
            headers=headers,
            data=params,
            method='POST'
        ):
            try:
                return api_helper.parse_auth_link(response.response_data)
            except IndexError:
                raise NoUrlFound(
                    'Не удалось найти ссылку для авторизации '
                    'в ответе от апи, проверьте client_id'
                )

    @classmethod
    async def get_access_token(
            cls,
            code: str,
            client_id: str,
            redirect_uri: str = 'https://example.com'
    ) -> str:
        """
        Метод для получения токена для запросов к YooMoney API

        :param code: временный код, который был получен в методе base_authorize
        :param client_id: идентификатор приложения, тип string
        :param redirect_uri: воронка, куда попадет временный код,
         который нужен для получения основного токена
        :return: YooMoney API TOKEN
        """
        headers = api_helper.parse_headers(content_json=True)
        params = {
            'code': code,
            'client_id': client_id,
            'grant_type': 'authorization_code',
            'redirect_uri': redirect_uri
        }
        async for response in RequestManager(
                without_context=True,
                messages=ERROR_CODE_NUMBERS,
                cache_time=DEFAULT_CACHE_TIME
        ).fast().fetch(
            url=BASE_YOOMONEY_URL + '/oauth/token',
            headers=headers,
            data=params,
            method='POST',
            get_json=True
        ):
            return response.response_data.get('access_token')

    async def revoke_api_token(self) -> Optional[Dict[str, bool]]:
        """
        Метод для отмены прав токена, при этом все его права тоже пропадают
        Документация:
        https://yoomoney.ru/docs/wallet/using-api/authorization/revoke-access-token
        """
        headers = self._auth_token(
            api_helper.parse_headers(auth=True)
        )
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/revoke',
                method='POST',
                headers=headers
        ):
            if response.ok:
                return {'success': True}

    async def account_info(self) -> AccountInfo:
        """
        Метод для получения информации об аккаунте пользователя
        Подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/account-info

        :return: объект AccountInfo
        """
        headers = self._auth_token(
            api_helper.parse_headers(
                **content_and_auth
            )
        )
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/account-info',
                headers=headers,
                method='POST'
        ):
            return AccountInfo.parse_raw(response.response_data)

    async def transactions(
            self,
            operation_types: Optional[
                Union[List[OperationType], Tuple[OperationType, ...]]
            ] = None,
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None,
            start_record: Optional[int] = None,
            records: int = 30,
            label: Optional[Union[str, int]] = None
    ) -> List[Operation]:
        """
        Подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/operation-history\n
        Метод позволяет просматривать историю операций (полностью или частично)
        в постраничном режиме.\n
        Записи истории выдаются в обратном хронологическом порядке:
        от последних к более ранним.\n
        Перечень типов операций, которые требуется отобразить.\n
        Возможные значения:
        DEPOSITION — пополнение счета (приход);\n
        PAYMENT — платежи со счета (расход);\n
        INCOMING(incoming-transfers-unaccepted) —
         непринятые входящие P2P-переводы любого типа.\n

        :param operation_types: Тип операций
        :param label: string.
         Отбор платежей по значению метки.
         Выбираются платежи, у которых указано заданное значение параметра
         label вызова request-payment.
        :param start_date: datetime	Вывести операции от момента времени
         (операции, равные start_date, или более поздние)
         Если параметр отсутствует, выводятся все операции.
        :param end_date: datetime	Вывести операции до момента времени
         (операции более ранние, чем end_date).
         Если параметр отсутствует, выводятся все операции.
        :param start_record: string
         Если параметр присутствует, то будут отображены операции, начиная
          с номера start_record.
         Операции нумеруются с 0. Подробнее про постраничный вывод списка
        :param records:	int	Количество запрашиваемых записей истории операций.
         Допустимые значения: от 1 до 100, по умолчанию — 30.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**content_and_auth)
        )

        if records <= 0 or records > 100:
            raise InvalidData(
                'Неверное количество записей. '
                'Кол-во записей, которые можно запросить,'
                ' находиться в диапазоне от 1 до 100 включительно'
            )

        data = {
            'records': records,
            'label': label
        }

        if operation_types:
            if all(
                    isinstance(
                        operation_type, OperationType
                    ) for operation_type in operation_types
            ):
                op_types = [
                    operation_type.value for operation_type in operation_types
                ]
                data.update({'type': ' '.join(
                    op_types)}
                )

        if isinstance(start_record, int):
            if start_record < 0:
                raise InvalidData('Укажите позитивное число')
            data.update({'start_record': start_record})

        if start_date:
            if not isinstance(start_date, datetime):
                raise InvalidData(
                    'Параметр start_date был передан неправильным типом данных'
                )
            data.update({'from': api_helper.datetime_to_str_in_iso(
                start_date, True
            )})

        if end_date:
            if not isinstance(end_date, datetime):
                raise InvalidData(
                    'Параметр end_date был передан неправильным типом данных'
                )
            data.update({'till': api_helper.datetime_to_str_in_iso(
                end_date, True
            )})

        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/operation-history',
                method='POST',
                headers=headers,
                data=self._requests.filter_dict(data),
                get_json=True
        ):
            return api_helper.multiply_objects_parse(
                lst_of_objects=response.response_data.get('operations'),
                model=Operation
            )

    async def transaction_info(self, operation_id: str) -> OperationDetails:
        """
        Позволяет получить детальную информацию об операции из истории.
        Требуемые права токена: operation-details.
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/user-account/operation-details

        :param operation_id: Идентификатор операции
        """
        headers = self._auth_token(api_helper.parse_headers(
            **content_and_auth)
        )
        payload = {
            'operation_id': operation_id
        }
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/operation-details',
                headers=headers,
                data=payload
        ):
            return OperationDetails.parse_raw(response.response_data)

    async def _pre_process_payment(
            self,
            to_account: str,
            amount: Union[int, float],
            pattern_id: str = 'p2p',
            comment_for_history: Optional[str] = None,
            comment_for_receiver: Optional[str] = None,
            protect: bool = False,
            expire_period: int = 1
    ) -> PreProcessPaymentResponse:
        """
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/request-payment\n
        Создание платежа, проверка параметров и возможности приема
        платежа магазином или перевода средств на счет пользователя ЮMoney.\n
        Данный метод не рекомендуется использовать напрямую, гораздо
        проще использовать send.
        Требуемые права токена:
        to-account («идентификатор получателя», «тип идентификатора»)

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода
         (номер счета, номер телефона или email).
        :param amount: Сумма к получению
         (придет на счет получателя счет после оплаты).
        :param comment_for_history: Комментарий к переводу,
         отображается в истории отправителя.
        :param comment_for_receiver:	string	Комментарий к переводу,
         отображается получателю.
        :param protect: Значение параметра true — признак того,
         что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции
            и получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365.
            Необязательный параметр. По умолчанию 1.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**content_and_auth))
        payload = {
            'pattern_id': pattern_id,
            'to': to_account,
            'amount_due': amount,
            'comment': comment_for_history,
            'message': comment_for_receiver,
            'expire_period': expire_period,
            'codepro': protect
        }
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/request-payment',
                headers=headers,
                data=payload
        ):
            try:
                return PreProcessPaymentResponse.parse_raw(
                    response.response_data
                )
            except ValidationError:
                msg = "Недостаточно денег для перевода или ошибка сервиса"
                self._requests.raise_exception(
                    status_code=400,
                    message=msg
                )

    async def send(
            self,
            to_account: str,
            amount: Union[int, float],
            money_source: str = 'wallet',
            pattern_id: str = 'p2p',
            cvv2_code: str = '',
            card_type: Optional[str] = None,
            protect: bool = False,
            comment_for_history: Optional[str] = None,
            comment: Optional[str] = None,
            expire_period: int = 1
    ) -> Payment:
        """
        Метод для отправки денег на аккаунт или карту другого человека.
        Данная функция делает сразу 2 запроса, из-за этого
         вы можете почувствовать небольшую потерю производительности,
          вы можете использовать метод
        _pre_process_payment и получать объект PreProcessPaymentResponse,
        в котором есть информация о ещё неподтвержденном платеже\n
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/process-payment

        :param pattern_id: Идентификатор шаблона платежа
        :param to_account: string Идентификатор получателя перевода
         (номер счета, номер телефона или email).
        :param amount: Сумма к получению
         (придет на счет получателя счет после оплаты). МИНИМАЛЬНАЯ СУММА 2.
        :param money_source: Запрашиваемый метод проведения платежа.
         wallet — со счета пользователя, если вы хотите использовать card,
         тогда нужно будет передать card_type
         для поиска карты в списке ваших банковских карт,
         а также опционально cvv2 код для проведения платежа
        :param comment_for_history: Комментарий к переводу,
         отображается в истории отправителя.
        :param card_type: Тип банковской карты, нужно заполнять,
         только если вы хотите списать средства с вашей карты
        :param cvv2_code: опционально, может быть не передан, однако,
         если для оплаты картой это требуется,
         параметр стоит передавать
        :param comment:	Комментарий к переводу, отображается получателю.
        :param protect: Значение параметра true — признак того,
         что перевод защищен кодом протекции.
         По умолчанию параметр отсутствует (обычный перевод).
        :param expire_period: Число дней, в течении которых:
            получатель перевода может ввести код протекции и
            получить перевод на свой счет,
            получатель перевода до востребования может получить перевод.
            Значение параметра должно находиться в интервале от 1 до 365.
            Необязательный параметр. По умолчанию 1.
        """
        if amount < 2:
            raise InvalidData(
                'Введите сумму, которая больше минимальной(2 и выше)'
            )
        pre_payment = await self._pre_process_payment(
            to_account=to_account,
            amount=amount,
            comment_for_history=comment_for_history,
            comment_for_receiver=comment,
            expire_period=expire_period,
            protect=protect,
            pattern_id=pattern_id
        )

        headers = self._auth_token(
            api_helper.parse_headers(**content_and_auth))

        payload = {
            'request_id': pre_payment.request_id,
            'money_source': 'wallet'
        }
        expression = isinstance(pre_payment, PreProcessPaymentResponse)
        if money_source == 'card' and expression:
            if pre_payment.money_source.cards.allowed == 'true':
                if not isinstance(card_type, str):
                    cards = pre_payment.money_source.cards
                    payload.update({
                        'money_source': cards.items[0].item_id,
                        'csc': cvv2_code
                    })
                else:
                    cards = pre_payment.money_source.cards.items
                    for card in cards:
                        if card.item_type == card_type:
                            payload.update(
                                {
                                    'money_source': card.item_id,
                                    'csc': cvv2_code
                                },
                            )
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/process-payment',
                method='POST',
                headers=headers,
                data=payload
        ):
            return Payment.parse_raw(
                response.response_data
            ).initialize(pre_payment.protection_code)

    async def get_balance(self) -> Optional[Union[float, int]]:
        """Метод для получения баланса на кошельке yoomoney"""
        return (await self.account_info()).balance

    async def accept_incoming_transaction(
            self,
            operation_id: str,
            protection_code: str
    ) -> IncomingTransaction:
        """
        Прием входящих переводов, защищенных кодом протекции,
        если вы передали в метод send параметр protect
        Количество попыток приема
        входящего перевода с кодом протекции ограничено.
        При исчерпании количества попыток, перевод автоматически отвергается
        (перевод возвращается отправителю).
        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-accept

        :param operation_id: Идентификатор операции,
         значение параметра operation_id ответа метода history()
        :param protection_code: Код протекции. Строка из 4-х десятичных цифр.
         Указывается для входящего перевода, защищенного кодом протекции.
         Для переводов до востребования отсутствует.
        """
        headers = self._auth_token(
            api_helper.parse_headers(**content_and_auth))
        payload = {
            'operation_id': operation_id,
            'protection_code': protection_code
        }
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/incoming-transfer-accept',
                headers=headers,
                data=payload,
                method='POST',
                get_json=True
        ):
            return IncomingTransaction.parse_raw(
                response.response_data
            )

    async def reject_transaction(self, operation_id: str) -> Dict[str, str]:
        """
        Отмена входящих переводов, защищенных кодом протекции, если вы передали
         в метод send параметр protect,
        и переводов до востребования. \n
        При отмене перевода он возвращается отправителю. \n
        Требуемые права токена: incoming-transfers

        Более подробная документация:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-reject

        :param operation_id: Идентификатор операции, значение параметра
         operation_id ответа метода history().
        :return: словарь в json формате с ответом от апи
        """
        headers = self._auth_token(
            api_helper.parse_headers(**content_and_auth))
        async for response in self._requests.fast().fetch(
                url=BASE_YOOMONEY_URL + '/api/incoming-transfer-reject',
                headers=headers,
                data={'operation_id': operation_id},
                method='POST',
                get_json=True
        ):
            return response.response_data

    async def check_transaction(
            self,
            amount: Union[int, float],
            transaction_type: str = 'in',
            comment: Optional[str] = None,
            rows_num: int = 100,
            sender_number: Optional[str] = None
    ) -> bool:
        """
        Метод для проверки транзакции.\n
        Данный метод использует self.transactions(rows_num=rows_num)
         для получения платежей.\n
        Для небольшой оптимизации вы можете уменьшить rows_num задав его,
        однако это не гарантирует правильный результат

        :param amount: сумма платежа
        :param transaction_type: тип платежа
        :param sender_number: номер получателя
        :param rows_num: кол-во платежей, которое будет проверяться
        :param comment: комментарий, по которому будет проверяться транзакция
        :return: bool, есть ли такая транзакция в истории платежей
        """
        # Generate types of transactions for request
        types = [
            OperationType.DEPOSITION if transaction_type == 'in' else OperationType.PAYMENT]
        # Get transactions by params
        transactions = await self.transactions(
            operation_types=types,
            records=rows_num
        )
        for txn in transactions:
            # Get details of transaction to check it later
            detail = await self.transaction_info(txn.operation_id)

            # Parse amount and comment,
            # because the parameters depend on the type of transaction
            amount_, comment_ = api_helper.parse_amount(
                transaction_type,
                detail
            )
            checked = api_helper.check_params(
                amount=amount,
                transaction_type=transaction_type,
                amount_=amount_,
                txn=detail
            )
            if checked:
                if txn.status == 'success':
                    if comment_ == comment and detail.sender == sender_number:
                        return True
                    elif isinstance(comment, str) and isinstance(
                            sender_number,
                            str):
                        continue
                    elif comment_ == comment:
                        return True

        return False
